
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <sys/malloc.h>

#include <machine/bus.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcidevs.h>

#include "elanvar.h"
#include "elan_mmcr.h"

struct device *elandev;

struct elan_softc {
	struct device		sc_dev;
	bus_space_tag_t		esc_mmcrt;
	bus_space_handle_t	esc_mmcr;
	bus_space_tag_t		esc_iot;
	bus_space_handle_t	esc_io;
};

#define ELAN_MMCR_BASE (0xfffef000)

#define ELAN_REVID_PRODUCT(r)	(((r) >> 8) & 0xff)
#define ELAN_REVID_MAJOR(r)	(((r) >> 4) & 0x0f)
#define ELAN_REVID_MINOR(r)	(((r) >> 0) & 0x0f)

#define ELAN_CPUCTL_CLOCK(r)	((r) & 0x3)
#define ELAN_CPUCTL_CLOCK_100MHZ 1
#define ELAN_CPUCTL_CLOCK_133MHZ 2

#define ELAN_CPUCTL_CACHE_WR_MODE (1 << 4)


static int elan_probe(struct device *, void *, void *);
static void elan_attach(struct device *, struct device *, void *);
static void elan_scan(struct device *, void *);
static int elan_print(void *aux, const char *pnp);
static int elan_submatch(struct device *parent, void *match, void *aux);

struct cfattach elan_ca = {
        sizeof(struct elan_softc), elan_probe, elan_attach
};

struct cfdriver elan_cd = {
        0, "elan", DV_DULL, 1
};

static int
elan_print(void *aux, const char *pnp)
{
	if(pnp)
		printf("elan at %s", pnp);

	return (UNCONF);
}

void
elan_init(struct device *parent, struct pci_attach_args *pa)
{
	struct elan_attach_args ea;

	memset(&ea, 0, sizeof ea);
	ea.elan_iot = pa->pa_iot;
	ea.elan_memt = pa->pa_memt;

	printf("\n");
	elandev = config_found_sm(parent, &ea, elan_print, elan_submatch);

	printf("\n");
	if(elandev) {
		config_scan(elan_scan, elandev);
	}
}

static int
elan_probe(struct device *parent, void *match, void *aux)
{
        struct cfdata *cf = match;

	return (cf->cf_attach == &elan_ca);
}

static void
elan_attach(struct device *parent, struct device *self, void *aux)
{
	struct elan_softc *sc = (struct elan_softc *)self;
	struct elan_attach_args *ea = aux;
	int s = splimp();
	u_int16_t tmp;

	sc->esc_mmcrt = ea->elan_memt;
	sc->esc_iot = ea->elan_iot;

	if(bus_space_map(sc->esc_mmcrt, ELAN_MMCR_BASE, 0x1000, 0, &sc->esc_mmcr)) {
		printf(": can't map MMCR\n");
		goto fail;
	}

#ifdef SOEKRIS_X1_FREQ
	/*
	 * At lesat one Soekris box has a 32.768MHz xtal instead of the
	 * supported 33.000 or 33.333 one.  It also leaves SWTMRCFG at
	 * 33.333MHz. Configuring for 33.000 is still wrong, but much
	 * closer to the truth.
	 */

#if SOEKRIS_X1_FREQ < ((33333333 + 33000000) / 2)
	bus_space_write_1(sc->esc_mmcrt, sc->esc_mmcr, ELAN_MMCR_SWTMRCFG, 1);
#else
	bus_space_write_1(sc->esc_mmcrt, sc->esc_mmcr, ELAN_MMCR_SWTMRCFG, 0);
#endif

#endif /* SOEKRIS_X1_FREQ */

	tmp = bus_space_read_2(sc->esc_mmcrt, sc->esc_mmcr, ELAN_MMCR_REVID);

	switch(ELAN_REVID_PRODUCT(tmp)) {
	case 0:
		printf(": AMD Elan SC520");
		break;
	default:
		printf(": Unknown product");
		break;
	}

	printf(" rev %d.%d", ELAN_REVID_MAJOR(tmp), ELAN_REVID_MINOR(tmp));

	tmp = bus_space_read_1(sc->esc_mmcrt, sc->esc_mmcr, ELAN_MMCR_CPUCTL);

	switch(ELAN_CPUCTL_CLOCK(tmp)) {
	case ELAN_CPUCTL_CLOCK_100MHZ:
		printf(" 100MHz");
		break;
	case ELAN_CPUCTL_CLOCK_133MHZ:
		printf(" 133MHz");
		break;
	default:
		printf(", *** invalid clock ***");
		break;
	}

	if(tmp & ELAN_CPUCTL_CACHE_WR_MODE) 
		printf(", write-back inhibited");

	printf(", reset=%b",
		bus_space_read_1(sc->esc_mmcrt, sc->esc_mmcr, ELAN_MMCR_RESSTA),
		"\20"
		"\x07" "SCP"
		"\x06" "HRST"
		"\x05" "SRST"
		"\x04" "WDT"
		"\x03" "SD"
		"\x02" "PRGRST"
		"\x01" "PWRGOOD");

	printf(", sysinfo=0x%2.2x",
		 bus_space_read_1(sc->esc_mmcrt, sc->esc_mmcr, ELAN_MMCR_SYSINFO));
	
fail:
	splx(s);
}

static int
elan_submatch(struct device *parent, void *match, void *aux)
{
        struct cfdata *cf = match;
        struct elan_attach_args *ma = aux;

	if(cf->cf_attach != &elan_ca)
		return (0);

        return ((*cf->cf_attach->ca_match)(parent, cf, ma));
}

static void
elan_scan(struct device *parent, void *match)
{
	struct elan_softc *sc = (struct elan_softc *)parent;
	struct device *dev = match;
	struct cfdata *cf = dev->dv_cfdata;
	struct elan_attach_args ea;

	memset(&ea, 0, sizeof ea);
	ea.elan_memt = sc->esc_mmcrt;
	ea.elan_mem = sc->esc_mmcr;
	ea.elan_iot = sc->esc_iot;
	ea.elan_io = sc->esc_io;

	if (autoconf_verbose)
		printf(">>> probing for %s%d\n", cf->cf_driver->cd_name,
			cf->cf_unit);

	if ((*cf->cf_attach->ca_match)(parent, dev, &ea) > 0) {
		if (autoconf_verbose)
			printf(">>> probing for %s%d succeeded\n",
				cf->cf_driver->cd_name, cf->cf_unit);
		config_attach(parent, dev, &ea, elan_print);
	} else {
		if (autoconf_verbose)
			printf(">>> probing for %s%d failed\n",
				cf->cf_driver->cd_name, cf->cf_unit);
		free(dev, M_DEVBUF);
	}
}

